/*++

Module Name:

    cse451p2.c

Abstract:

    This module implements the CSE451 Project2 NT Api monitoring calls

--*/

#include "exp.h"
#pragma hdrstop

#include "stdlib.h"
#include "string.h"
#include "vdmntos.h"
#include <nturtl.h>
#include "pool.h"
#include "stktrace.h"
#include "align.h"

//
//  This is the single lock we use to control access to our data structures
//

FAST_MUTEX CSE451Lock;

//
//  This struct contains the basic counters that we are keeping track of
//

SYSTEM_CSE451_INFORMATION CSE451ApiData;

//
//  We will buffer the history in at most 16 4KB chunnks referenced by this
//  array of pointers.  We know each entry to either add or return based on
//  a head and tail index.  The each index we can determine the chunk to
//  which it belongs and the index within the chunk.
//

PCSE451_API_EVENT CSE451Buffer[16];

ULONG CSE451BufferHead; // index to the next free spot in the buffer
ULONG CSE451BufferTail; // index of the next entry to return unless tail == head

#define EVENTS_PER_4K (4096/sizeof(CSE451_API_EVENT))

//
//  This is a local support routine to help add entries to the buffer
//

VOID
CSE451AddEntryToBuffer (
    CSE451_API_EVENT_TYPE Classification,
    OPTIONAL PNTSTATUS Status
    );


VOID
CSE451InitSystem (
    )

/*++

Routine Description:

    Initialize the counters and buffers used to do our API monitoring and history

Arguments:

    None

Return Value:

    None

--*/

{
    ULONG i;

    //
    //  Some useful debug prints but not needed on the running system
    //

    //DbgPrint("[%s:%d] CSE451InitSystem, sizeof(SYSTEM_CSE451_INFORMATION) = %d, sizeof(CSE451Buffer) = %d\n", __FILE__, __LINE__, sizeof(SYSTEM_CSE451_INFORMATION), sizeof(CSE451Buffer));
    //DbgPrint("Events per 4K = %d\n", EVENTS_PER_4K);

    //
    //  We start with zero counters
    //

    RtlZeroMemory( &CSE451ApiData, sizeof(SYSTEM_CSE451_INFORMATION) );

    //
    //  Initialize the lock
    //

    ExInitializeFastMutex(&CSE451Lock);

    //
    //  We start with empty buffer and no entries
    //

    CSE451BufferTail = CSE451BufferHead = 0;
    RtlZeroMemory( CSE451Buffer, sizeof(CSE451Buffer) );

    //
    //  And return to our caller
    //

    return;
}


NTSTATUS
CSE451ProcessQuery (
    PSYSTEM_CSE451_INFORMATION Buffer,
    ULONG BufferLength,
    OPTIONAL PULONG ReturnLength
    )

/*++

Routine Description:

    This routine processes the user's query request to return our collected
    API count information.

Arguments:

    None

Return Value:

    None

--*/

{
    ULONG BufferCapacity;
    ULONG TotalEvents;
    ULONG EventsToCopy;

    ULONG i;
    ULONG ChunkIndex;
    ULONG ChunkOffset;

    //
    // Make sure that the size is correct
    //

    if (BufferLength < sizeof( SYSTEM_CSE451_INFORMATION)) {

        return STATUS_INFO_LENGTH_MISMATCH;
    }

    //
    //  Acquire the global lock for our data structures
    //

    ExAcquireFastMutex( &CSE451Lock );

    //
    //  Copy over the counters before we tackle the event buffer
    //

    *Buffer = CSE451ApiData;

    //
    //  Now we can copy over the events from our saved buffer into the user's
    //  buffer.  See how much space the user has given us in terms of the number
    //  of events the buffer can hold, calculate the number of events we have to
    //  give to the caller and take the minimum of the two.
    //

    BufferCapacity = (BufferLength - FIELD_OFFSET(SYSTEM_CSE451_INFORMATION, Events)) / sizeof(CSE451_API_EVENT);
    TotalEvents = CSE451BufferHead - CSE451BufferTail;
    EventsToCopy = (BufferCapacity < TotalEvents ? BufferCapacity : TotalEvents);

    //DbgPrint("Tail = %d, Head = %d, Copy = %d\n", CSE451BufferTail, CSE451BufferHead, EventsToCopy);

    //
    //  Now we have the number of events to copy from our buffer to the
    //  user's buffer.  Instead of standing on our heads to give out what part
    //  of each buffer to copy over we'll do rather simplistic approach of
    //  copying each entry individually.
    //

    Buffer->NumberOfEvents = EventsToCopy;

    for (i = 0; i < EventsToCopy; i++) {

        //
        //  Compute the chunk and offset for the entry
        //

        ChunkIndex = (CSE451BufferTail + i) / EVENTS_PER_4K % 16;
        ChunkOffset = (CSE451BufferTail + i) % EVENTS_PER_4K;

        //
        //  Copy over the entry into the user buffer.  Note that we are still
        //  protected by the try except of our caller
        //

        Buffer->Events[i] = CSE451Buffer[ChunkIndex][ChunkOffset];

        //
        //  Now check if we can free the chunk.  We can free it if we've just crossed
        //  into a new 4KB boundary and the head is not in the chunk we've just left
        //

        if (((ChunkOffset+1) == EVENTS_PER_4K) && ((CSE451BufferHead / EVENTS_PER_4K % 16) != ChunkIndex)) {

            ExFreePool(CSE451Buffer[ChunkIndex]);
            CSE451Buffer[ChunkIndex] = NULL;
        }
    }

    //
    //  Update our tail
    //

    CSE451BufferTail += EventsToCopy;

    //
    //  Release our global lock and return to our caller, but the value we return
    //  will change if we actually copied over any events
    //

    ExReleaseFastMutex( &CSE451Lock );

    return (EventsToCopy != 0 ? STATUS_SUCCESS : STATUS_NO_MORE_ENTRIES);
}


/*++

    Routine Description:

        Each routine is responsible for acquiring and releasing the lock (mutex),
        incrementing the appropriate counter and calling the support routine
        to enqueue the buffer entry.

    Arguments:

        None for the Create calls and Status for the return calls

    Return Value:

        None

--*/

VOID CSE451CreateCall ()
{
    ExAcquireFastMutex( &CSE451Lock );
    CSE451ApiData.Create[0]++;
    CSE451AddEntryToBuffer( CreateCall, NULL );
    ExReleaseFastMutex( &CSE451Lock );
}

VOID CSE451CloseCall ()
{
    ExAcquireFastMutex( &CSE451Lock );
    CSE451ApiData.Close[0]++;
    CSE451AddEntryToBuffer( CloseCall, NULL );
    ExReleaseFastMutex( &CSE451Lock );
}

VOID CSE451OpenCall ()
{
    ExAcquireFastMutex( &CSE451Lock );
    CSE451ApiData.Open[0]++;
    CSE451AddEntryToBuffer( OpenCall, NULL );
    ExReleaseFastMutex( &CSE451Lock );
}

VOID CSE451ReadCall ()
{
    ExAcquireFastMutex( &CSE451Lock );
    CSE451ApiData.Read[0]++;
    CSE451AddEntryToBuffer( ReadCall, NULL );
    ExReleaseFastMutex( &CSE451Lock );
}

VOID CSE451WriteCall ()
{
    ExAcquireFastMutex( &CSE451Lock );
    CSE451ApiData.Write[0]++;
    CSE451AddEntryToBuffer( WriteCall, NULL );
    ExReleaseFastMutex( &CSE451Lock );
}

VOID CSE451GetCall ()
{
    ExAcquireFastMutex( &CSE451Lock );
    CSE451ApiData.Get[0]++;
    CSE451AddEntryToBuffer( GetCall, NULL );
    ExReleaseFastMutex( &CSE451Lock );
}

VOID CSE451SetCall ()
{
    ExAcquireFastMutex( &CSE451Lock );
    CSE451ApiData.Set[0]++;
    CSE451AddEntryToBuffer( SetCall, NULL );
    ExReleaseFastMutex( &CSE451Lock );
}

VOID CSE451QueryCall ()
{
    ExAcquireFastMutex( &CSE451Lock );
    CSE451ApiData.Query[0]++;
    CSE451AddEntryToBuffer( QueryCall, NULL );
    ExReleaseFastMutex( &CSE451Lock );
}

VOID
CSE451CreateReturn (
    IN NTSTATUS Status
    )
{
    ExAcquireFastMutex( &CSE451Lock );
    if (NT_INFORMATION(Status))  { CSE451ApiData.Create[2]++; }
    else if (NT_WARNING(Status)) { CSE451ApiData.Create[3]++; }
    else if (NT_ERROR(Status))   { CSE451ApiData.Create[4]++; }
    else                         { CSE451ApiData.Create[1]++; }
    CSE451AddEntryToBuffer( CreateReturn, &Status );
    ExReleaseFastMutex( &CSE451Lock );
}

VOID
CSE451CloseReturn (
    IN NTSTATUS Status
    )
{
    ExAcquireFastMutex( &CSE451Lock );
    if (NT_INFORMATION(Status))  { CSE451ApiData.Close[2]++; }
    else if (NT_WARNING(Status)) { CSE451ApiData.Close[3]++; }
    else if (NT_ERROR(Status))   { CSE451ApiData.Close[4]++; }
    else                         { CSE451ApiData.Close[1]++; }
    CSE451AddEntryToBuffer( CloseReturn, &Status );
    ExReleaseFastMutex( &CSE451Lock );
}

VOID
CSE451OpenReturn (
    IN NTSTATUS Status
    )
{
    ExAcquireFastMutex( &CSE451Lock );
    if (NT_INFORMATION(Status))  { CSE451ApiData.Open[2]++; }
    else if (NT_WARNING(Status)) { CSE451ApiData.Open[3]++; }
    else if (NT_ERROR(Status))   { CSE451ApiData.Open[4]++; }
    else                         { CSE451ApiData.Open[1]++; }
    CSE451AddEntryToBuffer( OpenReturn, &Status );
    ExReleaseFastMutex( &CSE451Lock );
}

VOID
CSE451ReadReturn (
    IN NTSTATUS Status
    )
{
    ExAcquireFastMutex( &CSE451Lock );
    if (NT_INFORMATION(Status))  { CSE451ApiData.Read[2]++; }
    else if (NT_WARNING(Status)) { CSE451ApiData.Read[3]++; }
    else if (NT_ERROR(Status))   { CSE451ApiData.Read[4]++; }
    else                         { CSE451ApiData.Read[1]++; }
    CSE451AddEntryToBuffer( ReadReturn, &Status );
    ExReleaseFastMutex( &CSE451Lock );
}

VOID
CSE451WriteReturn (
    IN NTSTATUS Status
    )
{
    ExAcquireFastMutex( &CSE451Lock );
    if (NT_INFORMATION(Status))  { CSE451ApiData.Write[2]++; }
    else if (NT_WARNING(Status)) { CSE451ApiData.Write[3]++; }
    else if (NT_ERROR(Status))   { CSE451ApiData.Write[4]++; }
    else                         { CSE451ApiData.Write[1]++; }
    CSE451AddEntryToBuffer( WriteReturn, &Status );
    ExReleaseFastMutex( &CSE451Lock );
}

VOID
CSE451GetReturn (
    IN NTSTATUS Status
    )
{
    ExAcquireFastMutex( &CSE451Lock );
    if (NT_INFORMATION(Status))  { CSE451ApiData.Get[2]++; }
    else if (NT_WARNING(Status)) { CSE451ApiData.Get[3]++; }
    else if (NT_ERROR(Status))   { CSE451ApiData.Get[4]++; }
    else                         { CSE451ApiData.Get[1]++; }
    CSE451AddEntryToBuffer( GetReturn, &Status );
    ExReleaseFastMutex( &CSE451Lock );
}

VOID
CSE451SetReturn (
    IN NTSTATUS Status
    )
{
    ExAcquireFastMutex( &CSE451Lock );
    if (NT_INFORMATION(Status))  { CSE451ApiData.Set[2]++; }
    else if (NT_WARNING(Status)) { CSE451ApiData.Set[3]++; }
    else if (NT_ERROR(Status))   { CSE451ApiData.Set[4]++; }
    else                         { CSE451ApiData.Set[1]++; }
    CSE451AddEntryToBuffer( SetReturn, &Status );
    ExReleaseFastMutex( &CSE451Lock );
}

VOID
CSE451QueryReturn (
    IN NTSTATUS Status
    )
{
    ExAcquireFastMutex( &CSE451Lock );
    if (NT_INFORMATION(Status))  { CSE451ApiData.Query[2]++; }
    else if (NT_WARNING(Status)) { CSE451ApiData.Query[3]++; }
    else if (NT_ERROR(Status))   { CSE451ApiData.Query[4]++; }
    else                         { CSE451ApiData.Query[1]++; }
    CSE451AddEntryToBuffer( QueryReturn, &Status );
    ExReleaseFastMutex( &CSE451Lock );
}


//
//  Local Support Routine
//

VOID
CSE451AddEntryToBuffer (
    CSE451_API_EVENT_TYPE Classification,
    OPTIONAL PNTSTATUS Status
    )

/*++

Routine Description:

    This routine adds a new entry to the API history buffer

Arguments:

    Classification - The type of entry we are adding to the history

    Status - Pointer to an optional status to associate with the entry

Return Value:

    None

--*/

{
    LARGE_INTEGER Time;

    ULONG ChunkIndex;
    ULONG ChunkOffset;

    //
    //  First check if the buffer is full.  It is full if the head is up to
    //  64KB ahead of the tail
    //

    if (CSE451BufferHead >= CSE451BufferTail + (EVENTS_PER_4K*16)) {

        //
        //  Buffer overflow!  So munge the previous entry that we added to
        //  indicate overflow and return to our caller
        //

        ChunkIndex = (CSE451BufferHead-1) / EVENTS_PER_4K % 16;
        ChunkOffset = (CSE451BufferHead-1) % EVENTS_PER_4K;
        CSE451Buffer[ChunkIndex][ChunkOffset].EventType = BufferOverflow;

        return;
    }

    //
    //  We have space to add an entry to compute the chunk and offset where
    //  to put it
    //

    ChunkIndex = (CSE451BufferHead) / EVENTS_PER_4K % 16;
    ChunkOffset = (CSE451BufferHead) % EVENTS_PER_4K;

    //
    //  Get our system time
    //

    KeQuerySystemTime(&Time);

    //
    //  If the chunk where we want to put the value is null then allocate new
    //  chunk for us to use
    //

    if (CSE451Buffer[ChunkIndex] == NULL) {

        CSE451Buffer[ChunkIndex] = ExAllocatePool(NonPagedPool, 4096);
    }

    //
    //  We have a chunk, so now add entry and increment our head
    //

    CSE451Buffer[ChunkIndex][ChunkOffset].Time = Time;
    CSE451Buffer[ChunkIndex][ChunkOffset].EventType = Classification;
    CSE451Buffer[ChunkIndex][ChunkOffset].Status = (Status == NULL ? 0 : *Status);

    CSE451BufferHead += 1;

    //
    //  And return to our caller
    //

    return;
}

